import sys
from walt.server.images.image import NodeImage
from walt.server.network import nfs
from walt.common.tools import \
        failsafe_makedirs, failsafe_symlink
from walt.server.images.image import get_mount_path
# About terminology: See comment about it in image.py.

MSG_IMAGE_IS_USED_BUT_NOT_FOUND=\
    "WARNING: image %s is not found. Cannot attach it to related nodes.\n"
CONFIG_ITEM_DEFAULT_IMAGE='default_image'
MSG_WOULD_OVERWRITE_IMAGE="""\
An image has the same name in your working set.
This operation would overwrite it%s.
"""
MSG_WOULD_OVERWRITE_IMAGE_REBOOTED_NODES='\
 (and reboot %d node(s))'

class NodeImageStore(object):
    def __init__(self, docker, db):
        self.docker = docker
        self.db = db
        self.images = {}
    def refresh(self):
        db_images = { db_img.fullname: db_img.ready \
                        for db_img in self.db.select('images') }
        # add missing images
        for db_fullname in db_images:
            if '/walt-node' in db_fullname and db_fullname not in self.images:
                self.images[db_fullname] = NodeImage(self.docker, db_fullname)
            # and update their 'ready' status
            image = self.images[db_fullname]
            image.set_ready(db_images[db_fullname])
        # remove deleted images
        for fullname in self.images.keys():
            if fullname not in db_images:
                del self.images[fullname]
    def register_image(self, image_fullname, is_ready):
        self.db.insert('images', fullname=image_fullname, ready=is_ready)
        self.db.commit()
        self.refresh()
    def rename(self, old_fullname, new_fullname):
        self.db.execute('update images set fullname = %(new_fullname)s where fullname = %(old_fullname)s',
                            dict(old_fullname = old_fullname, new_fullname = new_fullname))
        self.db.commit()
        self.refresh()
    def remove(self, image_fullname):
        self.db.delete('images', fullname=image_fullname)
        self.db.commit()
        self.refresh()
    def set_image_ready(self, image_fullname):
        self.db.update('images', 'fullname', fullname=image_fullname, ready=True)
        self.db.commit()
        self.refresh()
    def __getitem__(self, image_fullname):
        return self.images[image_fullname]
    def __iter__(self):
        return self.images.iterkeys()
    def __len__(self):
        return len(self.images)
    def keys(self):
        return self.images.keys()
    def iteritems(self):
        return self.images.iteritems()
    def values(self):
        return self.images.values()
    # look for an image belonging to the requester.
    # The 'expected' parameter allows to specify if we expect a matching
    # result (expected = True), no matching result (expected = False),
    # or if both options are ok (expected = None).
    # If expected is True or False and the result does not match expectation,
    # an error message will be printed.
    def get_user_image_from_tag(self, requester, image_tag, expected = True, ready_only = True):
        found = None
        for image in self.images.values():
            if image.tag == image_tag and image.user == requester.username:
                found = image
        if expected == True and found is None:
            requester.stderr.write(
                "Error: No such image '%s'. (tip: walt image show)\n" % image_tag)
        if expected == False and found is not None:
            requester.stderr.write(
                "Error: Image '%s' already exists.\n" % image_tag)
        if expected == True and found is not None:
            if ready_only and found.ready == False:
                requester.stderr.write(
                    "Error: Image '%s' is not ready.\n" % image_tag)
                found = None
        return found
    def get_user_unmounted_image_from_tag(self, requester, image_tag):
        image = self.get_user_image_from_tag(requester, image_tag)
        if image:   # otherwise issue is already reported
            if image.mounted:
                requester.stderr.write('Sorry, cannot proceed because the image is mounted.\n')
                return None
        return image
    def update_image_mounts(self, images_in_use = None):
        if images_in_use == None:
            images_in_use = self.get_images_in_use()
        images_found = []
        # ensure all needed images are mounted
        for fullname in images_in_use:
            if fullname in self.images:
                img = self.images[fullname]
                if not img.mounted:
                    img.mount()
                images_found.append(img)
            else:
                sys.stderr.write(MSG_IMAGE_IS_USED_BUT_NOT_FOUND % fullname)
        # update nfs configuration
        nfs.update_exported_filesystems(images_found)
        # unmount images that are not needed anymore
        for fullname in self.images:
            if fullname not in images_in_use:
                img = self.images[fullname]
                if img.mounted:
                    img.unmount()
    def cleanup(self):
        # release nfs mounts
        nfs.update_exported_filesystems([])
        # unmount images
        for fullname in self.images:
            img = self.images[fullname]
            if img.mounted:
                img.unmount()
    def get_images_in_use(self):
        res = set([ item.image for item in \
            self.db.execute("""
                SELECT DISTINCT image FROM nodes""").fetchall()])
        return res
    def get_default_image(self, node_type):
        return 'waltplatform/walt-node:%s-default' % node_type
    def umount_used_image(self, image):
        images = self.get_images_in_use()
        images.remove(image.fullname)
        self.update_image_mounts(images)
    def warn_overwrite_image(self, requester, image_fullname):
        num_nodes = len(self.db.select("nodes", image=image_fullname))
        if num_nodes == 0:
            reboot_message = ''
        else:
            reboot_message = MSG_WOULD_OVERWRITE_IMAGE_REBOOTED_NODES % num_nodes
        requester.stderr.write(MSG_WOULD_OVERWRITE_IMAGE % reboot_message)




#!/usr/bin/env python

from walt.server.devices.topology import Topology
from walt.server import snmp
import re

DEVICE_NAME_NOT_FOUND="""No device with name '%s' found.\n"""

NEW_NAME_ERROR_AND_GUIDELINES = """\
Failed: invalid new name.
This name must be a valid network hostname.
Only alphanumeric characters and '-' are allowed.
The 1st character must be an alphabet letter.
(Example: rpi-D327)
"""

class DevicesManager(object):

    def __init__(self, db):
        self.db = db
        self.topology = Topology(db)

    def rescan(self, **kwargs):
        self.topology.rescan(**kwargs)

    def rename(self, requester, old_name, new_name):
        device_info = self.get_device_info(requester, old_name)
        if device_info == None:
            return
        if self.get_device_info(requester, new_name, err_message = None) != None:
            requester.stderr.write("""Failed: a device with name '%s' already exists.\n""" % new_name)
            return
        if not re.match('^[a-zA-Z][a-zA-Z0-9-]*$', new_name):
            requester.stderr.write(NEW_NAME_ERROR_AND_GUIDELINES)
            return
        # all is fine, let's update it
        self.db.update("devices", 'mac', mac = device_info.mac, name = new_name)
        self.db.commit()

    def get_device_info(self, requester, device_name, \
                        err_message = DEVICE_NAME_NOT_FOUND):
        device_info = self.db.select_unique("devices", name=device_name)
        if device_info == None and err_message != None:
            requester.stderr.write(err_message % device_name)
        return device_info

    def get_name_from_ip(self, ip):
        device_info = self.db.select_unique("devices", ip=ip)
        return device_info.name

    def notify_unknown_ip(self, requester, device_name):
        requester.stderr.write('Sorry, IP address of %s in unknown.\n' \
                            % device_name)

    def get_device_ip(self, requester, device_name):
        device_info = self.get_device_info(requester, device_name)
        if device_info == None:
            return None # error already reported
        if device_info.ip == None:
            self.notify_unknown_ip(requester, device_name)
        return device_info.ip

    def add(self, **kwargs):
        self.topology.add_device(**kwargs)

    def is_reachable(self, requester, device_name):
        res = self.get_device_info(requester, device_name)
        if res == None:
            return
        return res.reachable != 0

    def node_bootup_event(self, node_ip):
        # if we got this event, then the node is reachable
        self.db.update('devices', 'ip', ip=node_ip, reachable=1)
        self.db.commit()



#!/usr/bin/env python
from base import load_mib

class IPSetupProxy(object):
    def __init__(self, snmp_proxy):
        self.snmp = snmp_proxy
        load_mib("NETGEAR-SWITCHING-MIB")

    def perform_dhcp_setup(self):
        # the switch is already configured to boot using DHCP
        # by default, but affecting this value again causes
        # the switch to restart the DHCP procedure, which is
        # exactly what we expect.
        self.snmp.agentNetworkConfigProtocol = 3 # dhcp

    def record_current_ip_config_as_static(self):
        # if server and switches are restarted, the switches may
        # send a DHCP request before the DHCP server of the WalT server
        # is started.
        # This causes the switches to choose a default address,
        # e.g. 192.168.0.239 for Netgear switches.
        # This causes major issues because several switches
        # may get this same address.
        # Thus, the first time a switch is detected with a DHCP IP belonging
        # to the WalT network, we statically set this IP in its bootup
        # procedure.
        current_ip = str(self.snmp.agentNetworkIPAddress)
        current_netmask = str(self.snmp.agentNetworkSubnetMask)
        with self.snmp as batch:
            batch.agentNetworkConfigProtocol = 1
            batch.agentNetworkIPAddress = current_ip
            batch.agentNetworkSubnetMask = current_netmask



#!/usr/bin/env python
import os, pty, shlex, uuid, fcntl, termios, sys
from subprocess import Popen, PIPE, STDOUT
from walt.common.io import SmartBufferingFileReader, \
                            unbuffered, read_and_copy
from walt.common.tcp import Requests, read_pickle

class ForkPtyProcessListener(object):
    def __init__(self, slave_r, slave_w, sock_file):
        self.slave_r = slave_r
        self.slave_w = slave_w
        self.sock_file = sock_file
        self.slave_reader = SmartBufferingFileReader(slave_r)
    # let the event loop know what we are reading on
    def fileno(self):
        return self.slave_r.fileno()
    # when the event loop detects an event for us, this means
    # the slave proces wrote something on its output
    # we just have to copy this to the user socket
    def handle_event(self, ts):
        return read_and_copy(
                self.slave_reader, self.sock_file)
    def close(self):
        # note: if we close, we want the other listener
        # (ParallelProcessSocketListener) to close too.
        # that's why we kept a reference to sock_file_r
        # and slave_w, closing them should achieve what
        # we want.
        self.sock_file.close()
        self.slave_r.close()
        self.slave_w.close()

class ParallelProcessSocketListener(object):
    def __init__(self, ev_loop, sock_file, **kwargs):
        self.ev_loop = ev_loop
        self.params = None
        self.sock_file = sock_file
        # provide read_available() method on the socket
        self.sock_reader = SmartBufferingFileReader(self.sock_file)
        self.slave_r = None
        self.slave_w = None
        self.sock_file.write('READY\n')
    def update_params(self):
        pass  # override in subclasses if needed
    def get_command(self, **params):
        '''this should be defined in subclasses'''
        raise NotImplementedError
    def start(self):
        cmd_args = shlex.split(self.params['cmd'])
        env = os.environ.copy()
        if 'env' in self.params:
            env.update(self.params['env'])
        if 'pty' in self.params and self.params['pty']:
            self.start_pty(cmd_args, env)
        else:
            self.start_popen(cmd_args, env)
    def start_pty(self, cmd_args, env):
        # fork a child process in its own virtual terminal
        slave_pid, fd_slave = pty.fork()
        # the child (slave process) should execute the command
        if slave_pid == 0:
            # set the window size appropriate for the remote client
            if 'win_size' in self.params:
                fcntl.ioctl(sys.stdout.fileno(), termios.TIOCSWINSZ,
                                self.params['win_size'])
            os.execvpe(cmd_args[0], cmd_args, env)
        # the parent should communicate.
        # use unbuffered communication with the slave process
        self.slave_r = os.fdopen(os.dup(fd_slave), 'r', 0)
        self.slave_w = os.fdopen(os.dup(fd_slave), 'w', 0)
        # create a new listener on the event loop for reading
        # what the slave process outputs
        process_listener = ForkPtyProcessListener(
                    self.slave_r, self.slave_w,
                    self.sock_file)
        self.ev_loop.register_listener(process_listener)
    def start_popen(self, cmd_args, env):
        # For efficiency, we let the popen object read directly from the socket.
        # Thus the ev_loop should not longer detect input data on this socket,
        # it should only detect errors, that is why we call update_listener() below.
        popen = Popen(cmd_args, env=env, bufsize=1024*1024,
                        stdin=self.sock_file, stdout=self.sock_file, stderr=STDOUT)
        self.ev_loop.update_listener(self, 0)
    # let the event loop know what we are reading on
    def fileno(self):
        if self.sock_file.closed:
            return None
        return self.sock_file.fileno()
    # handle_event() will be called when the event loop detects
    # new input data for us.
    def handle_event(self, ts):
        if self.params == None:
            # we did not get the parameters yet, let's do it
            self.params = read_pickle(self.sock_file)
            if self.params == None:
                self.close()    # issue
                return
            self.update_params()
            self.params['cmd'] = self.get_command(**self.params)
            # we now have all info to start the child process
            self.start()
        else:
            # otherwise we are all set. Thus, getting input data means
            # the user wrote something on the prompt (i.e. the socket)
            # we just have to copy this to the slave process input
            return read_and_copy(
                self.sock_reader, self.slave_w)
    def close(self):
        self.sock_file.close()
        if self.slave_r:
            self.slave_r.close()
        if self.slave_w:
            self.slave_w.close()





from walt.common.tcp import Requests
from walt.server.parallel import ParallelProcessSocketListener
from walt.server.const import SSH_COMMAND
import os

TYPE_CLIENT = 0
TYPE_IMAGE = 1

HELP_INVALID = """\
Usage:
$ walt $(image_or_node_label) cp <local_file_path> <$(image_or_node_label)>:<file_path>
or
$ walt $(image_or_node_label) cp <$(image_or_node_label)>:<file_path> <local_file_path>

Regular files as well as directories are accepted.
"""

def analyse_file_types(requester, image_tag_or_node, src_path, src_fs, dst_path, dst_fs, **kwargs):
    bad = dict(valid = False)
    dst_dir = None
    src_type = src_fs.get_file_type(src_path)
    dst_type = dst_fs.get_file_type(dst_path)
    if dst_type is None:
        # maybe this is just the target filename, let's verify that the parent
        # directory exists
        parent_path = os.path.dirname(dst_path)
        if dst_fs.get_file_type(parent_path) == 'd':
            # ok
            dst_type = 'd'
            dst_name = os.path.basename(dst_path)
            dst_dir = parent_path
    for ftype, path in [(src_type, src_path), (dst_type, dst_path)]:
        if ftype is None:
            requester.stderr.write('No such file or directory: %s\n' % path)
            return bad
    if dst_type == 'f':
        if src_type == 'd':
            requester.stderr.write(
                "Invalid request. " + \
                "Overwriting regular file %s with directory %s is not allowed.\n" % \
                    (dst_path, src_path))
            return bad
        # overwriting a file
        dst_type = 'd'
        dst_name = os.path.basename(dst_path)
        dst_dir = os.path.dirname(dst_path)
    elif dst_dir is None:
        # copying to a directory, keeping the source name
        dst_name = os.path.basename(src_path)
        dst_dir = dst_path
    kwargs.update(
        valid = True,
        dst_dir = dst_dir,
        dst_name = dst_name
    )
    return kwargs

def validate_cp(image_or_node_label, caller,
                requester, src, dst):
    invalid = False
    operands = []
    operand_index_per_type = {}
    filesystems = []
    paths = []
    image_tag = None
    for index, operand in enumerate([src, dst]):
        parts = operand.split(':')
        operand_type = len(parts)-1
        if operand_type > 1:
            invalid = True
            break
        operands.append(operand)
        operand_index_per_type[operand_type] = index
        if operand_type == TYPE_CLIENT:
            filesystems.append(requester.filesystem)
            paths.append(operand.rstrip('/'))
        else:
            image_tag_or_node, path = parts
            if not caller.validate_cp_entity(requester, image_tag_or_node):
                return
            filesystem = caller.get_cp_entity_filesystem(
                                    requester, image_tag_or_node)
            if not filesystem.ping():
                requester.stderr.write(\
                    "Could not reach %s. Try again later.\n" % image_tag_or_node)
                return
            filesystems.append(filesystem)
            paths.append(path.rstrip('/'))
    if len(operand_index_per_type) != 2:
        invalid = True
    if invalid:
        requester.stderr.write(HELP_INVALID % dict(
            image_or_node_label = image_or_node_label
        ))
        return
    src_fs, dst_fs = filesystems
    src_path, dst_path = [
            path if path.startswith('/') else './' + path
            for path in paths ]
    info = analyse_file_types(  requester, image_tag_or_node,
                                src_path, src_fs,
                                dst_path, dst_fs)
    if info.pop('valid') == False:
        return
    # all seems fine
    client_operand_index = operand_index_per_type[TYPE_CLIENT]
    info.update(
        src_dir = os.path.dirname(src_path),
        src_name = os.path.basename(src_path),
        client_operand_index = client_operand_index,
        **caller.get_cp_entity_attrs(requester, image_tag_or_node)
    )
    # return an immutable object
    return tuple(info.items())

def docker_wrap_cmd(cmd):
    cmd_args = cmd.split()
    return '''\
        docker run -i --name %%(container_name)s --entrypoint %s \
        %%(image_fullname)s %s ''' % (
            cmd_args[0],
            ' '.join(cmd_args[1:])
        )

def ssh_wrap_cmd(cmd):
    return SSH_COMMAND + ' root@%(node_ip)s "' + cmd + '"'

TarSendCommand='''\
        tar c -C %(src_dir)s %(src_name)s \
        --transform 's/^[^\/]*/%(dst_name)s/' '''

TarReceiveCommand='''\
        tar x -C %(dst_dir)s'''

class ImageTarSender(ParallelProcessSocketListener):
    REQ_ID = Requests.REQ_TAR_FROM_IMAGE
    def get_command(self, **params):
        return docker_wrap_cmd(TarSendCommand) % params

class ImageTarReceiver(ParallelProcessSocketListener):
    REQ_ID = Requests.REQ_TAR_TO_IMAGE
    def get_command(self, **params):
        return docker_wrap_cmd(TarReceiveCommand) % params

class NodeTarSender(ParallelProcessSocketListener):
    REQ_ID = Requests.REQ_TAR_FROM_NODE
    def get_command(self, **params):
        return ssh_wrap_cmd(TarSendCommand) % params

class NodeTarReceiver(ParallelProcessSocketListener):
    REQ_ID = Requests.REQ_TAR_TO_NODE
    def get_command(self, **params):
        return ssh_wrap_cmd(TarReceiveCommand) % params

class TransferManager(object):
    def __init__(self, tcp_server, ev_loop):
        for cls in [    ImageTarSender,
                        ImageTarReceiver,
                        NodeTarSender,
                        NodeTarReceiver ]:
            tcp_server.register_listener_class(
                    req_id = cls.REQ_ID,
                    cls = cls,
                    ev_loop = ev_loop)




import socket, re, cPickle as pickle
from datetime import datetime
from walt.common.tcp import read_pickle, write_pickle, \
                            Requests

class LogsToDBHandler(object):
    def __init__(self, db):
        self.db = db

    def log(self, **record):
        self.db.insert('logs', **record)

class LogsHub(object):
    def __init__(self):
        self.handlers = set([])

    def addHandler(self, handler):
        self.handlers.add(handler)

    def removeHandler(self, handler):
        self.handlers.remove(handler)

    def log(self, **kwargs):
        to_be_removed = set([])
        for handler in self.handlers:
            res = handler.log(**kwargs)
            # a handler may request to be deleted
            # by returning False
            if res == False:
                to_be_removed.add(handler)
        for handler in to_be_removed:
            self.handlers.remove(handler)

class LogsStreamListener(object):
    def __init__(self, db, hub, sock, sock_file, **kwargs):
        self.db = db
        self.hub = hub
        self.sock = sock
        self.sock_file = sock_file
        self.stream_id = None

    def register_stream(self):
        name = str(read_pickle(self.sock_file))
        sender_ip, sender_port = self.sock.getpeername()
        sender_info = self.db.select_unique('devices', ip = sender_ip)
        if sender_info == None:
            sender_mac = None
        else:
            sender_mac = sender_info.mac
        stream_info = self.db.select_unique('logstreams',
                            sender_mac = sender_mac, name = name)
        if stream_info:
            # found existing stream
            stream_id = stream_info.id
        else:
            # register new stream
            stream_id = self.db.insert('logstreams', returning='id',
                            sender_mac = sender_mac, name = name)
        # these are not needed anymore
        self.db = None
        self.sock = None
        return stream_id

    # let the event loop know what we are reading on
    def fileno(self):
        return self.sock_file.fileno()
    # when the event loop detects an event for us, we
    # know a log line should be read. 
    def handle_event(self, ts):
        if self.stream_id == None:
            self.stream_id = self.register_stream()
            # register_stream() involves a read on the stream
            # to get its name.
            # supposedly that's why we have been woken up.
            # let the event loop call us again if there is more.
            return True
        record = read_pickle(self.sock_file)
        if record == None:
            print 'Log stream with id %d is being closed.' % self.stream_id
            # let the event loop know we should 
            # be removed.
            return False
        # convert timestamp to datetime
        ts = record['timestamp']
        if not isinstance(ts, datetime):
            record['timestamp'] = datetime.fromtimestamp(ts)
        record.update(stream_id=self.stream_id)
        self.hub.log(**record)
        return True
    def close(self):
        self.sock_file.close()

class RetrieveDBLogsTask(object):
    def __init__(self, client_handler):
        self.handler = client_handler
        self.db = self.handler.db
    def prepare(self):
        # ensure all past logs are commited
        self.db.commit()
        # create a server cursor
        self.c = self.db.get_server_cursor()
    def perform(self):
        # this is where things may take some time...
        for record in self.db.get_logs(self.c, **self.handler.params):
            d = record._asdict()
            if self.handler.write_to_client(
                        senders_filtered = True, **d) == False:
                break
        del self.c
    def handle_result(self, res):
        self.handler.notify_history_processed()

class LogsToSocketHandler(object):
    def __init__(self, db, hub, sock, sock_file, blocking, **kwargs):
        self.db = db
        self.db_logs_task = RetrieveDBLogsTask(self)
        self.sock_file_r = sock_file
        self.sock_file_w = sock.makefile('w', 0)
        self.cache = {}
        self.params = None
        self.hub = hub
        self.blocking = blocking
        self.retrieving_from_db = None
        self.realtime_buffer = []
        #sock.settimeout(1.0)
    def log(self, **record):
        # if we are processing the history part,
        # do not send the realtime logs right now
        if self.retrieving_from_db:
            self.realtime_buffer.append(record)
        else:   # ok, write to client
            return self.write_to_client(**record)
    def notify_history_processed(self):
        if self.params['realtime']:
            # done with the history part.
            # we can flush the buffer of realtime logs
            for record in self.realtime_buffer:
                if self.write_to_client(**record) == False:
                    break
            # notify that next logs can be sent
            # directly to the client
            self.retrieving_from_db = False
        else:
            # no realtime mode, we can quit
            self.close()
    def write_to_client(self, stream_id, senders_filtered=False, **record):
        try:
            if stream_id not in self.cache:
                self.cache[stream_id] = self.db.execute(
                """SELECT d.name as sender, s.name as stream
                   FROM logstreams s, devices d
                   WHERE s.id = %s
                     AND s.sender_mac = d.mac
                """ % stream_id).fetchall()[0]._asdict()
            stream_info = self.cache[stream_id]
            # when data comes from the db, senders are already filtered,
            # while data coming from the hub has to be filtered.
            if not senders_filtered:
                if stream_info['sender'] not in self.params['senders']:
                    return  # filter out
            # matching the streams is always done here, otherwise there
            # may be inconsistencies between the regexp format in the postgresql
            # database and in python
            if self.streams_regexp:
                matches = self.streams_regexp.findall(stream_info['stream'])
                if len(matches) == 0:
                    return  # filter out
            d = {}
            d.update(record)
            d.update(stream_info)
            if self.sock_file_w.closed:
                raise IOError()
            write_pickle(d, self.sock_file_w)
        except IOError as e:
            # the socket was supposedly closed.
            print "client log connection closing"
            # notify the hub that we should be removed.
            return False
    # let the event loop know what we are reading on
    def fileno(self):
        return self.sock_file_r.fileno()
    # this is what we will do depending on the client request params
    def handle_params(self, history, realtime, senders, streams):
        if history:
            # unpickle the elements of the history range
            history = (pickle.loads(e) if e else None for e in history)
        if streams:
            self.streams_regexp = re.compile(streams)
        else:
            self.streams_regexp = None
        self.params = dict( history = history,
                            realtime = realtime,
                            senders = senders)
        if history:
            self.retrieving_from_db = True
            self.db_logs_task.prepare()
            self.blocking.do(self.db_logs_task)
        else:
            self.retrieving_from_db = False
        if realtime:
            self.hub.addHandler(self)
    # this is what we do when the event loop detects an event for us
    def handle_event(self, ts):
        if self.params == None:
            params = read_pickle(self.sock_file_r)
            self.handle_params(**params)
        else:
            return False    # no more communication is expected this way
    def close(self):
        self.sock_file_r.close()
        self.sock_file_w.close()

class LogsManager(object):
    def __init__(self, db, tcp_server, blocking):
        self.db = db
        self.blocking = blocking
        self.hub = LogsHub()
        self.hub.addHandler(LogsToDBHandler(db))
        tcp_server.register_listener_class(
                    req_id = Requests.REQ_DUMP_LOGS,
                    cls = LogsToSocketHandler,
                    db = self.db,
                    hub = self.hub,
                    blocking = self.blocking)
        tcp_server.register_listener_class(
                    req_id = Requests.REQ_NEW_INCOMING_LOGS,
                    cls = LogsStreamListener,
                    db = self.db,
                    hub = self.hub)

    def get_checkpoint(self, requester, cp_name, expected=True):
        cp_info = self.db.select_unique(
            'checkpoints', name=cp_name, username=requester.username)
        if expected and cp_info == None:
            requester.stderr.write("Failed: no checkpoint with this name '%s'.\n" % cp_name)
        if not expected and cp_info != None:
            requester.stderr.write('Failed: a checkpoint with this name already exists.\n')
        return cp_info

    def add_checkpoint(self, requester, cp_name, date):
        if self.get_checkpoint(requester, cp_name, expected=False) != None:
            return
        if not date:
            date = datetime.now()
        self.db.insert('checkpoints',
                name=cp_name, username=requester.username, timestamp=date)
        requester.stdout.write("New checkpoint %s recorded at server time: %s.\n" % (cp_name, date))

    def remove_checkpoint(self, requester, cp_name):
        if self.get_checkpoint(requester, cp_name, expected=True) == None:
            return
        self.db.delete('checkpoints', name=cp_name, username=requester.username)
        requester.stdout.write("Done.\n")

    def list_checkpoints(self, requester):
        res = self.db.select('checkpoints', username=requester.username)
        if len(res) == 0:
            requester.stdout.write('You own no checkpoints.\n')
        else:
            # re-execute because we don't want the 'username' column.
            requester.stdout.write(
                self.db.pretty_printed_select("""
                    SELECT timestamp, name FROM checkpoints
                    WHERE username = %s;
            """, (requester.username,)) + '\n')

    def get_pickled_checkpoint_time(self, requester, cp_name):
        cp_info = self.get_checkpoint(requester, cp_name, expected=True)
        if cp_info == None:
            return
        return pickle.dumps(cp_info.timestamp)




#!/usr/bin/env python
import select, subprocess, shlex
from ipaddress import ip_address, ip_network
from walt.common.tools import do, succeeds
from walt.server import const
from walt.server.snmp.proxy import Proxy

def ip(ip_as_str):
    return ip_address(unicode(ip_as_str))

def net(net_as_str):
    return ip_network(unicode(net_as_str))

def find_free_ip_near(ip, intf, increment):
    target_ip = ip
    while True:
        target_ip += increment
        if succeeds('arping -D -w 1 -I %s %s' % (intf, target_ip)):
            return target_ip

def smallest_subnet_for_these_ip_addresses(ip1, ip2):
    # start with <ip1>/31, then <ip1>/30 etc.
    # until <ip2> is in this network too.
    for netrange in range(31,0,-1):
        net = ip_network(u'%s/%d' % (ip1, netrange), strict=False)
        if ip2 in net:
            return net

def get_mac_address(intf):
    with open('/sys/class/net/' + intf +'/address') as f:
        return f.read().strip()

def dhcp_wait_ip(intf, ui, msg, explain=None, todo=None):
    # dhclient will go to background when an IP is obtained,
    # which should release the popen process.
    ui.task_start(msg, explain=explain, todo=todo)
    cmd = 'dhclient -1 %s' % intf
    while True:
        dh_client = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE)
        poller = select.poll()
        poller.register(dh_client.stdout, select.POLLIN)
        while len(poller.poll(500)) == 0:   # timeout unit is milliseconds
            ui.task_running()
        poller.unregister(dh_client.stdout)
        # check the return code, if not 0 loop again
        dh_client.wait()
        if dh_client.returncode == 0:
            break
    ui.task_done()

def dhcp_stop():
    do('dhclient -r')

def add_ip_to_interface(ip, subnet, intf):
    do('ip addr add %s/%d dev %s' % (ip, subnet.prefixlen, intf))

def del_ip_from_interface(ip, subnet, intf):
    do('ip addr del %s/%d dev %s' % (ip, subnet.prefixlen, intf))

def check_if_we_can_reach(remote_ip):
    return succeeds('ping -c 1 -w 1 %s' % remote_ip)

def is_walt_address(ip):
    return ip in net(const.WALT_SUBNET)

def assign_temp_ip_to_reach_neighbor(neighbor_ip, callback, intf, *args):
    reached = False
    callback_result = None
    for increment in [ 1, -1 ]:
        free_ip = find_free_ip_near(neighbor_ip, intf, increment)
        subnet = smallest_subnet_for_these_ip_addresses(neighbor_ip, free_ip)
        print free_ip, subnet
        add_ip_to_interface(free_ip, subnet, intf)
        if check_if_we_can_reach(neighbor_ip):
            callback_result = callback(free_ip, neighbor_ip, intf, *args)
            reached = True
        del_ip_from_interface(free_ip, subnet, intf)
        if reached:
            break
    return (reached, callback_result)

def restart_dhcp_setup_on_switch_cb(local_ip, switch_ip, intf):
    p = Proxy(str(switch_ip), ipsetup=True)
    p.ipsetup.perform_dhcp_setup()

def restart_dhcp_setup_on_switch(switch_ip):
    reached, res = assign_temp_ip_to_reach_neighbor(
                                ip(switch_ip),
                                restart_dhcp_setup_on_switch_cb,
                                const.WALT_INTF)
    if not reached:
        print 'Warning: Could not reach %s and restart its DHCP client.' % switch_ip

def set_static_ip_on_switch(switch_ip):
    p = Proxy(switch_ip, ipsetup=True)
    p.ipsetup.record_current_ip_config_as_static()

def lldp_update():
    do('lldpcli update')

def get_server_ip():
    subnet = net(const.WALT_SUBNET)
    return list(subnet.hosts()).pop(0)

def set_server_ip():
    do('ip link set up dev %s' % const.WALT_INTF)
    add_ip_to_interface(
            get_server_ip(),
            net(const.WALT_SUBNET),
            const.WALT_INTF)
    # let neighbors know we have updated things
    lldp_update()

def ip_in_walt_network(input_ip):
    subnet = net(const.WALT_SUBNET)
    return ip(input_ip) in subnet




from walt.server.tools import format_paragraph, columnate

NODE_SHOW_QUERY = """
    SELECT  d.name as name, d.type as type,
        split_part(n.image, '/', 1) as image_owner,
        split_part(n.image, ':', 2) as image_tag,
        i.ready as image_ready,
        d.ip as ip,
        (case when d.reachable = 1 then 'yes' else 'NO' end) as reachable
    FROM devices d, nodes n, images i
    WHERE   d.mac = n.mac
    AND     n.image = i.fullname
    ORDER BY image_owner, name;"""

MSG_USING_NO_NODES = """\
You are currently using no nodes. (tip: walt --help-about node-terminology)"""

MSG_RERUN_WITH_ALL = """\
Re-run with --all to see all deployable nodes."""

TITLE_NODE_SHOW_FREE_NODES_PART = """\
The following nodes are free (they have never been used):"""

TITLE_NODE_SHOW_USER_NODES_PART = """\
The following nodes are running one of your images:"""

TITLE_NODE_SHOW_OTHER_NODES_PART = """\
The following nodes are likely to be used by other users, since you do
not own the image deployed on them."""

TITLE_NODE_SHOW_NOT_READY_NODES_PART = """\
The following nodes were detected but are not available for now:
the startup OS image they will boot is being downloaded."""

MSG_NO_NODES = """\
No nodes detected!"""

MSG_NO_OTHER_NODES = """\
No other nodes were detected (apart from the ones listed above)."""

def show(db, requester, show_all):
    res_user, res_free, res_other, res_not_ready = [], [], [], []
    res = db.execute(NODE_SHOW_QUERY)
    for record in res:
        if not record.image_ready:
            res_not_ready.append(record)
        else:
            if record.image_owner == requester.username:
                res_user.append(record)
            elif record.image_owner == 'waltplatform':
                res_free.append(record)
            else:
                res_other.append(record)
    result_msg = ''
    if len(res_user) + len(res_free) == 0 and not show_all:
        return MSG_USING_NO_NODES + '\n' + MSG_RERUN_WITH_ALL
    if len(res_user) > 0:
        # display nodes of requester
        footnote = None
        if not show_all and len(res_free) == 0:
            footnote = MSG_RERUN_WITH_ALL
        table = [ (record.name, record.type,
                   record.image_tag, record.ip, record.reachable) \
                    for record in res_user ]
        header = [ 'name', 'type', 'image', 'ip', 'reachable' ]
        result_msg += format_paragraph(
                        TITLE_NODE_SHOW_USER_NODES_PART,
                        columnate(table, header=header),
                        footnote)
    if len(res_free) > 0:
        # display free nodes
        footnote = None
        if not show_all:
            footnote = MSG_RERUN_WITH_ALL
        table = [ (record.name, record.type, record.ip, record.reachable) \
                    for record in res_free ]
        header = [ 'name', 'type', 'ip', 'reachable' ]
        result_msg += format_paragraph(
                        TITLE_NODE_SHOW_FREE_NODES_PART,
                        columnate(table, header=header),
                        footnote)
    if not show_all:
        return result_msg
    if len(res_other) + len(res_user) + len(res_free) + len(res_not_ready) == 0:
        return MSG_NO_NODES + '\n'
    if len(res_other) + len(res_not_ready) == 0:
        result_msg += MSG_NO_OTHER_NODES + '\n'
    else:
        if len(res_other) > 0:
            # display nodes of other users
            table = [  (record.name, record.type, record.image_owner,
                        'server:%s/%s' % (record.image_owner, record.image_tag),
                        record.ip, record.reachable) \
                            for record in res_other ]
            header = [ 'name', 'type', 'image_owner', 'clonable_image_link', 'ip', 'reachable' ]
            result_msg += format_paragraph(
                        TITLE_NODE_SHOW_OTHER_NODES_PART,
                        columnate(table, header=header))
        if len(res_not_ready) > 0:
            # display nodes whose image is currently being downloaded
            table = [  (record.name, record.type, record.ip) \
                            for record in res_not_ready ]
            header = [ 'name', 'type', 'ip' ]
            result_msg += format_paragraph(
                        TITLE_NODE_SHOW_NOT_READY_NODES_PART,
                        columnate(table, header=header))
    return result_msg




from walt.server.images.image import validate_image_tag

def do_rename(images, docker, old_fullname, new_tag):
    name, old_tag = old_fullname.split(':')
    new_fullname = "%s:%s" % (name, new_tag)
    # rename the docker image
    docker.tag(old_fullname, new_fullname)
    docker.rmi(old_fullname)
    # update the store
    images.rename(old_fullname, new_fullname)

def rename(images, docker, requester, image_tag, new_tag):
    if not validate_image_tag(requester, new_tag):
        return
    image = images.get_user_unmounted_image_from_tag(requester, image_tag)
    if image:   # otherwise issue is already reported
        if not images.get_user_image_from_tag(requester, new_tag, expected=False):
            do_rename(images, docker, image.fullname, new_tag)




import os, sys, select, cPickle as pickle
from walt.server.const import UI_FIFO_PATH, UI_RESPONSE_FIFO_PATH
from walt.common.tools import failsafe_mkfifo

class UIManager(object):
    def ui_running(self):
        return os.path.exists(UI_FIFO_PATH)
    def send_request_to_ui(self, *args):
        with open(UI_FIFO_PATH, 'w') as fifo:
            pickle.dump(args, fifo)
    def wait_user_keypress(self):
        if self.ui_running():
            failsafe_mkfifo(UI_RESPONSE_FIFO_PATH)
            response_fifo = os.fdopen(
                os.open(UI_RESPONSE_FIFO_PATH, os.O_RDWR | os.O_NONBLOCK), 'r', 0)
            self.send_request_to_ui('WAIT_ENTER', UI_RESPONSE_FIFO_PATH)
            # block until we get the response message back
            poller = select.poll()
            poller.register(response_fifo, select.POLLIN)
            poller.poll()
            poller.unregister(response_fifo)
            pickle.load(response_fifo)
            response_fifo.close()
            os.remove(UI_RESPONSE_FIFO_PATH)
        else:
            print "Press <enter> to continue...",
            raw_input()
    def request_ui_update(self, *args):
        if self.ui_running():
            self.send_request_to_ui(*args)
            return True
        else:
            return False
    def task_start(self, msg, explain=None, todo=None):
        self.task_idx = 0
        self.task_msg = msg
        self.task_explain = explain
        self.task_todo = todo
        self.task_explained_ui = False
        self.set_status(msg)
    def task_running(self, activity_sign=True):
        if self.task_explain and not self.task_explained_ui:
            self.task_explained_ui = self.set_explain(
                self.task_explain,
                self.task_todo,
                ui_only = True
            )
        if activity_sign:
            status_text = "%s %s" % \
                (self.task_msg, '|/-\\'[self.task_idx])
            self.task_idx = (self.task_idx + 1)%4
        else:
            status_text = self.task_msg
        self.set_status(status_text, ui_only=True)
    def task_done(self):
        status_text = "%s %s" % (self.task_msg, 'done')
        self.set_status(status_text)
        self.request_ui_update('EXPLAIN', '')
    def task_failed(self, error_msg):
        status_text = "%s %s" % (self.task_msg, 'FAILED!!')
        self.set_status(status_text)
        self.set_explain('ERROR:\n' + error_msg, None)
    def update_topic(self, topic, text, ui_only=False):
        if not ui_only:
            print '**', text
        return self.request_ui_update(topic, text)
    def set_status(self, *args, **kwargs):
        return self.update_topic('STATUS', *args, **kwargs)
    def set_explain(self, explain, todo, **kwargs):
        succeeded = self.update_topic('EXPLAIN', explain, **kwargs)
        if succeeded and todo is not None:
            succeeded = self.update_topic('TODO', todo, **kwargs)
        return succeeded




from multiprocessing import Pipe
from threading import Thread
STOP = -1

class BlockingTasksThread(Thread):
    def __init__(self, tasks, pipe_in, pipe_out):
        Thread.__init__(self)
        self.tasks = tasks
        self.pipe_in = pipe_in
        self.pipe_out = pipe_out

    def run(self):
        while True:
            msg = self.pipe_in.recv()
            if msg == STOP:
                break
            task_id = msg
            task = self.tasks[task_id]
            try:
                task.result = task.perform()
            except Exception as e:
                task.result = e
            self.pipe_out.send(task_id)
        self.pipe_in.close()
        self.pipe_out.close()

class BlockingTasksManager(object):
    def __init__(self):
        self.next_task_id = 0
        self.tasks = {}
        self.pipe_requests, pipe_requests_child = Pipe()
        self.pipe_done, pipe_done_child = Pipe()
        self.thread = BlockingTasksThread(
                    self.tasks, pipe_requests_child, pipe_done_child)
        self.thread.start()

    def join_event_loop(self, ev_loop):
        self.ev_loop = ev_loop
        ev_loop.register_listener(self)

    def do(self, blocking_task):
        task_id = self.next_task_id
        self.next_task_id += 1
        self.tasks[task_id] = blocking_task
        self.pipe_requests.send(task_id)

    # let the event loop know what we are reading on
    def fileno(self):
        return self.pipe_done.fileno()

    # when the event loop detects an event for us, this
    # means the background process has completed a
    # task.
    def handle_event(self, ts):
        task_id = self.pipe_done.recv()
        task = self.tasks[task_id]
        task.handle_result(task.result)
        del self.tasks[task_id]

    def close(self):
        self.pipe_requests.send(STOP)
        self.thread.join()
        self.pipe_requests.close()
        self.pipe_done.close()

    def cleanup(self):
        self.close()




#!/usr/bin/env python
from walt.common.tcp import Requests
from walt.server.parallel import ParallelProcessSocketListener
from walt.server.const import SSH_COMMAND

INTERACTIVE_SSH_COMMAND = SSH_COMMAND + ' -tq'

class PromptSocketListener(ParallelProcessSocketListener):
    def update_params(self):
        self.params.update(
            env={'TERM':'xterm'},
            pty=True    # start subprocess in a virtual terminal
        )

class SQLPromptSocketListener(PromptSocketListener):
    REQ_ID = Requests.REQ_SQL_PROMPT
    def get_command(self, **params):
        return 'psql walt'

class DockerPromptSocketListener(PromptSocketListener):
    REQ_ID = Requests.REQ_DOCKER_PROMPT
    def get_command(self, **params):
        return 'docker run -it --entrypoint %s -h %s --name %s %s' % \
                       ('/bin/bash', 'image-shell',
                        params['container_name'], params['image_fullname'])

class NodeCmdSocketListener(PromptSocketListener):
    REQ_ID = Requests.REQ_NODE_CMD
    def get_command(self, node_ip, cmdargs, **kwargs):
        cmd = ' '.join(cmdargs)
        quoted_cmd = "'" + "'\"'\"'".join(cmd.split("'")) + "'"
        return '%(ssh)s root@%(ip)s %(cmd)s' % dict(
            ssh = INTERACTIVE_SSH_COMMAND,
            ip = node_ip,
            cmd = quoted_cmd
        )

class DevicePingSocketListener(PromptSocketListener):
    REQ_ID = Requests.REQ_DEVICE_PING
    def get_command(self, **params):
        return 'ping %s' % params['device_ip']

class InteractionManager(object):
    def __init__(self, tcp_server, ev_loop):
        for cls in [    SQLPromptSocketListener,
                        DockerPromptSocketListener,
                        NodeCmdSocketListener,
                        DevicePingSocketListener ]:
            tcp_server.register_listener_class(
                    req_id = cls.REQ_ID,
                    cls = cls,
                    ev_loop = ev_loop)




#!/usr/bin/env python

from walt.common.tools import get_mac_address
from walt.common.nodetypes import get_node_type_from_mac_address
from walt.server.network.tools import ip_in_walt_network, lldp_update, \
                                        restart_dhcp_setup_on_switch, \
                                        set_static_ip_on_switch
from walt.server.tools import format_paragraph, columnate
from walt.server.tree import Tree
from walt.server import snmp, const
import time, re

TOPOLOGY_QUERY = """
    SELECT  d1.name as name, d1.type as type, d1.mac as mac,
            d1.ip as ip,
            (case when d1.reachable = 1 then 'yes' else 'NO' end) as reachable,
            d2.name as switch_name, t.switch_port as switch_port
    FROM devices d1, devices d2, topology t
    WHERE   d1.mac = t.mac and d2.mac = t.switch_mac
    UNION ALL
    SELECT  d1.name as name, d1.type as type, d1.mac as mac,
            d1.ip as ip,
            (case when d1.reachable = 1 then 'yes' else 'NO' end) as reachable,
            NULL as switch_name, NULL as switch_port
    FROM devices d1, topology t
    WHERE   d1.mac = t.mac and t.switch_mac is null
    ORDER BY switch_name, switch_port;"""

FLOATING_DEVICES_QUERY = """
    SELECT  d1.name as name, d1.type as type, d1.mac as mac,
            d1.ip as ip,
            (case when d1.reachable = 1 then 'yes' else 'NO' end) as reachable
    FROM devices d1 LEFT JOIN topology t
    ON   d1.mac = t.mac
    WHERE t.mac is NULL;"""

MSG_DEVICE_TREE_EXPLAIN_UNREACHABLE = """\
note: devices marked with parentheses are unreachable
"""

MSG_DEVICE_TREE_MORE_DETAILS = """
tips:
- use 'walt device rescan' to update
- use 'walt device show' for more info
"""

TITLE_DEVICE_SHOW_MAIN = """\
The WalT network contains the following devices:"""

FOOTNOTE_DEVICE_SHOW_MAIN = """\
tips:
- use 'walt device tree' for a tree view of the network
- use 'walt device rescan' to update
- use 'walt device forget <device_name>' in case of a broken device"""

TITLE_DEVICE_SHOW_FLOATING = """\
The network position of the following devices is unknown (at least for now):"""

FOOTNOTE_DEVICE_SHOW_FLOATING = """\
tips:
- use 'walt device rescan' in a few minutes to update
- use 'walt device forget <device_name>' to make WalT forget about an obsolete device"""

class Topology(object):

    def __init__(self, db):
        self.db = db

    def get_type(self, mac):
        node_type = get_node_type_from_mac_address(mac)
        if node_type != None:
            # this is a node
            return node_type.SHORT_NAME
        elif mac == self.server_mac:
            return 'server'
        else:
            return 'switch'

    def collect_connected_devices(self, ui, host, host_is_a_switch,
                            host_mac, processed_switches):

        print "collect devices connected on %s" % host
        switches_with_dhcp_restarted = set()
        # avoid to loop forever...
        if host_is_a_switch:
            processed_switches.add(host_mac)
        while True:
            issue = False
            # get a SNMP proxy with LLDP feature
            snmp_proxy = snmp.Proxy(host, lldp=True)

            # record neighbors in db and recurse
            for port, neighbor_info in snmp_proxy.lldp.get_neighbors().items():
                ip, mac = neighbor_info['ip'], neighbor_info['mac']
                device_type = self.get_type(mac)
                if mac in processed_switches:
                    continue
                if not ip_in_walt_network(ip):
                    # if someone is viewing the UI, let him know that we are still
                    # alive, because this could cause a long delay.
                    if ui:
                        ui.task_running()
                    print 'Not ready, one neighbor has ip %s (not in WalT network yet)...' % ip
                    if device_type == 'switch' and mac not in switches_with_dhcp_restarted:
                        print 'trying to restart the dhcp client on switch %s (%s)' % (ip, mac)
                        switches_with_dhcp_restarted.add(mac)
                        restart_dhcp_setup_on_switch(ip)
                    lldp_update()
                    time.sleep(1)
                    issue = True
                    break
                if host_is_a_switch:
                    switch_mac, switch_port = host_mac, port
                else:
                    switch_mac, switch_port = None, None
                device_is_new = self.add_device(type=device_type,
                                mac=mac,
                                switch_mac=switch_mac,
                                switch_port=switch_port,
                                ip=ip)
                if device_type == 'switch':
                    if device_is_new:
                        print 'Affecting static IP configuration on switch %s (%s)...' % \
                            (ip, mac)
                        set_static_ip_on_switch(ip)
                    # recursively discover devices connected to this switch
                    self.collect_connected_devices(ui, ip, True,
                                            mac, processed_switches)
            if not issue:
                break   # otherwise restart the loop

    def rescan(self, requester=None, ui=None):
        self.db.execute("""
            UPDATE devices
            SET reachable = 0;""")

        self.server_mac = get_mac_address(const.WALT_INTF)
        self.collect_connected_devices(ui, "localhost", False, self.server_mac, set())
        self.db.commit()

        if requester != None:
            requester.stdout.write('done.\n')

    def generate_device_name(self, **kwargs):
        if kwargs['type'] == 'server':
            return 'walt-server'
        return "%s-%s" %(
            kwargs['type'],
            "".join(kwargs['mac'].split(':')[3:]))

    def add_device(self, **kwargs):
        # if we are there then we can reach this device
        kwargs['reachable'] = 1
        # update device info
        num_rows = self.db.update("devices", 'mac', **kwargs)
        # if no row was updated, this is a new device
        if num_rows == 0:
            # generate a name for this device
            name = self.generate_device_name(**kwargs)
            kwargs['name'] = name
            # insert a new row
            self.db.insert("devices", **kwargs)
        if 'switch_mac' in kwargs:
            mac = kwargs['mac']
            switch_mac = kwargs['switch_mac']
            switch_port = kwargs['switch_port']
            # add/update topology info:
            # - update or insert topology info for device with specified mac
            # - if any other device was connected at this place, it will
            #   be replaced (thus this other device will appear disconnected)
            self.db.delete('topology', mac = mac)
            self.db.delete('topology', switch_mac = switch_mac,
                                       switch_port = switch_port)
            self.db.insert("topology", **kwargs)
        return num_rows == 0 # return True if new device

    def tree(self):
        t = Tree()
        unreachable_found = False
        for device in self.db.execute(TOPOLOGY_QUERY).fetchall():
            name = device.name
            if device.reachable == 'NO':
                unreachable_found = True
                name = '(%s)' % name
            swport = device.switch_port
            if swport == None:
                label = name
                # align to 2nd letter of the name
                subtree_offset = 1
            else:
                label = '%d: %s' % (swport, name)
                # align to 2nd letter of the name
                subtree_offset = label.find(' ') + 2
            parent_key = device.switch_name
            t.add_node( name,   # will be the key in the tree
                        label,
                        subtree_offset=subtree_offset,
                        parent_key = parent_key)
        note = MSG_DEVICE_TREE_MORE_DETAILS
        if unreachable_found:
            note += MSG_DEVICE_TREE_EXPLAIN_UNREACHABLE
        return "\n%s%s" % (t.printed(), note)

    def show(self):
        # ** message about connected devices
        msg = format_paragraph(
                TITLE_DEVICE_SHOW_MAIN,
                self.db.pretty_printed_select(TOPOLOGY_QUERY),
                FOOTNOTE_DEVICE_SHOW_MAIN)
        # ** message about floating devices, if at least one
        res = self.db.execute(FLOATING_DEVICES_QUERY).fetchall()
        if len(res) > 0:
            msg += format_paragraph(
                        TITLE_DEVICE_SHOW_FLOATING,
                        self.db.pretty_printed_resultset(res),
                        FOOTNOTE_DEVICE_SHOW_FLOATING)
        return msg

    def setpower(self, device_mac, poweron):
        # we have to know on which PoE switch port the node is
        switch_ip, switch_port = self.get_connectivity_info(device_mac)
        if not switch_ip:
            return False
        # if powering off, the device will be unreachable
        if not poweron:
            self.db.update('devices', 'mac', mac=device_mac, reachable=0)
            self.db.commit()
        # let's request the switch to enable or disable the PoE
        proxy = snmp.Proxy(switch_ip, poe=True)
        proxy.poe.set_port(switch_port, poweron)
        return True

    def get_connectivity_info(self, device_mac):
        topology_info = self.db.select_unique("topology", mac=device_mac)
        if not topology_info:
            return (None, None)
        switch_mac = topology_info.switch_mac
        switch_port = topology_info.switch_port
        switch_info = self.db.select_unique("devices", mac=switch_mac)
        return switch_info.ip, switch_port



#!/usr/bin/env python

import rpyc, sys, datetime, cPickle as pickle
import walt.server as server
from walt.server.network import setup
from walt.server.network.tools import set_server_ip
from walt.server.tools import AutoCleaner
from walt.server.ui.manager import UIManager
from walt.common.daemon import WalTDaemon
from walt.common.constants import           \
                 WALT_SERVER_DAEMON_PORT,   \
                 WALT_NODE_DAEMON_PORT

WALT_SERVER_DAEMON_VERSION = 0.1

class ClientMirroringService(rpyc.Service):
    services_per_node = {}

    def on_connect(self):
        node_id = id(self._conn.root)
        self._node_id = node_id
        ClientMirroringService.services_per_node[node_id] = self

    def on_disconnect(self):
        del ClientMirroringService.services_per_node[self._node_id]

    def __del__(self):
        self._client = None

    def register_client(self, client):
        self._client = client

    @staticmethod
    def link_node_to_client(node, client):
        service = ClientMirroringService.services_per_node[id(node)]
        service.register_client(client)

    # forward all other method accesses to self._client
    def __getattr__(self, attr_name):
        return getattr(self._client, attr_name)

class ServerToNodeLink:
    def __init__(self, ip_address, client = None):
        self.node_ip = ip_address
        self.client = client

    def __enter__(self):
        if self.client:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT,
                            service = ClientMirroringService)
            node_service = self.conn.root
            ClientMirroringService.link_node_to_client(node_service, self.client)
        else:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT)
        return self.conn.root

    def __exit__(self, type, value, traceback):
        self.conn.close()

class PlatformService(rpyc.Service):
    ALIASES=("WalT_Platform",)

    def __init__(self, *args, **kwargs):
        rpyc.Service.__init__(self, *args, **kwargs)
        self.server = server.instance
        self.images = server.instance.images
        self.devices = server.instance.devices
        self.nodes = server.instance.nodes
        self.logs = server.instance.logs

    def on_connect(self):
        self._client = self._conn.root

    def on_disconnect(self):
        self._client = None

    def exposed_device_rescan(self):
        self.server.device_rescan(self._client)

    def exposed_device_tree(self):
        return self.devices.topology.tree()

    def exposed_device_show(self):
        return self.devices.topology.show()

    def exposed_show_nodes(self, show_all):
        return self.nodes.show(self._client, show_all)

    def exposed_get_reachable_nodes_ip(self, node_set):
        return self.nodes.get_reachable_nodes_ip(
                        self._client, node_set)

    def exposed_get_device_ip(self, device_name):
        return self.devices.get_device_ip(
                        self._client, device_name)

    def exposed_get_node_ip(self, node_name):
        return self.nodes.get_node_ip(
                        self._client, node_name)

    def exposed_blink(self, node_name, blink_status):
        node_ip = self.nodes.get_reachable_node_ip(
                        self._client, node_name)
        if node_ip == None:
            return False # error was already reported
        with ServerToNodeLink(node_ip, self._client) as node_service:
            node_service.blink(blink_status)
        return True

    def exposed_parse_set_of_nodes(self, node_set):
        nodes = self.nodes.parse_node_set(self._client, node_set)
        if nodes:
            return [n.name for n in nodes]

    def exposed_includes_nodes_not_owned(self, node_set, warn):
        return self.nodes.includes_nodes_not_owned(self._client, node_set, warn)

    def exposed_poweroff(self, node_set, warn_unknown_topology):
        return self.nodes.setpower(self._client, node_set, False, warn_unknown_topology)

    def exposed_poweron(self, node_set, warn_unknown_topology):
        return self.nodes.setpower(self._client, node_set, True, warn_unknown_topology)

    def exposed_validate_node_cp(self, src, dst):
        return self.nodes.validate_cp(self._client, src, dst)

    def exposed_wait_for_nodes(self, q, node_set):
        self.nodes.wait(self._client, q, node_set)

    def exposed_rename(self, old_name, new_name):
        self.server.rename_device(self._client, old_name, new_name)

    def exposed_has_image(self, image_tag):
        return self.images.has_image(self._client, image_tag)

    def exposed_set_image(self, node_set, image_tag, warn_unknown_topology):
        self.server.set_image(self._client, node_set, image_tag, warn_unknown_topology)

    def exposed_is_device_reachable(self, device_name):
        return self.devices.is_reachable(self._client, device_name)

    def exposed_count_logs(self, history, **kwargs):
        unpickled_history = (pickle.loads(e) if e else None for e in history)
        return self.server.db.count_logs(history = unpickled_history, **kwargs)

    def exposed_forget(self, device_name):
        self.server.forget_device(device_name)

    def exposed_fix_image_owner(self, other_user):
        return self.images.fix_owner(self._client, other_user)

    def exposed_search_images(self, q, keyword):
        self.images.search(self._client, q, keyword)

    def exposed_clone_image(self, q, clonable_link, force=False):
        self.images.clone(requester = self._client,
                          q = q,
                          clonable_link = clonable_link,
                          force = force)

    def exposed_show_images(self):
        return self.images.show(self._client.username)

    def exposed_create_image_shell_session(self, image_tag):
        return self.images.create_shell_session(self._client, image_tag)

    def exposed_remove_image(self, image_tag):
        self.images.remove(self._client, image_tag)

    def exposed_rename_image(self, image_tag, new_tag):
        self.images.rename(self._client, image_tag, new_tag)

    def exposed_duplicate_image(self, image_tag, new_tag):
        self.images.duplicate(self._client, image_tag, new_tag)

    def exposed_validate_image_cp(self, src, dst):
        return self.images.validate_cp(self._client, src, dst)

    def exposed_node_bootup_event(self):
        node_ip, node_port = self._conn._config['endpoints'][1]
        self.devices.node_bootup_event(node_ip)
        node_name = self.devices.get_name_from_ip(node_ip)
        self.nodes.node_bootup_event(node_name)

    def exposed_add_checkpoint(self, cp_name, pickled_date):
        date = None
        if pickled_date:
            date = pickle.loads(pickled_date)
        self.logs.add_checkpoint(self._client, cp_name, date)

    def exposed_remove_checkpoint(self, cp_name):
        self.logs.remove_checkpoint(self._client, cp_name)

    def exposed_list_checkpoints(self):
        self.logs.list_checkpoints(self._client)

    def exposed_get_pickled_time(self):
        return pickle.dumps(datetime.datetime.now())

    def exposed_get_pickled_checkpoint_time(self, cp_name):
        return self.logs.get_pickled_checkpoint_time(self._client, cp_name)

class WalTServerDaemon(WalTDaemon):
    """WalT (wireless testbed) server daemon."""
    VERSION = WALT_SERVER_DAEMON_VERSION

    def getParameters(self):
        return dict(
                service_cl = PlatformService,
                port = WALT_SERVER_DAEMON_PORT,
                ev_loop = server.instance.ev_loop)

    def init_end(self):
        server.instance.ui.set_status('Ready.')

def run():
    ui = UIManager()
    myserver = server.Server(ui)
    # set ip on WalT network (eth0.1)
    set_server_ip()
    myserver.dhcpd.update(force=True)
    setup.setup(ui)
    with AutoCleaner(myserver) as server.instance:
        myserver.update()
        WalTServerDaemon.run()

if __name__ == "__main__":
    run()




from walt.common.tcp import Requests
from walt.server.nodes.register import NodeRegistrationHandler
from walt.server.nodes.show import show
from walt.server.nodes.wait import WaitInfo
from walt.server.tools import format_sentence_about_nodes, \
                                merge_named_tuples
from walt.common.nodetypes import is_a_node_type_name
from walt.server.transfer import validate_cp
from walt.server.filesystem import Filesystem
from walt.server.const import SSH_COMMAND

NODE_SET_QUERIES = {
        'my-nodes': """
            SELECT  d.name as name
            FROM devices d, nodes n, images i
            WHERE   d.mac = n.mac
            AND     n.image = i.fullname
            AND     i.ready = True
            AND     split_part(n.image, '/', 1) = '%s'
            ORDER BY name;""",
        'all-nodes': """
            SELECT  d.name as name
            FROM devices d, nodes n, images i
            WHERE   d.mac = n.mac
            AND     n.image = i.fullname
            AND     i.ready = True
            ORDER BY name;"""
}

MSG_NODE_NEVER_REGISTERED = """\
Node %s was eventually detected but never registered itself to the server.
It seems it is not running a valid WalT boot image.
"""

MSG_CONNECTIVITY_UNKNOWN = """\
%s: Unknown PoE switch port. Cannot proceed! Please retry in a few minutes.
"""

FS_CMD_PATTERN = SSH_COMMAND + ' root@%(node_ip)s %%(prog)s %%(prog_args)s'

class NodesManager(object):
    def __init__(self, db, tcp_server, devices, **kwargs):
        self.db = db
        self.current_register_requests = set()
        self.devices = devices
        self.kwargs = kwargs
        self.wait_info = WaitInfo()
        tcp_server.register_listener_class(
                    req_id = Requests.REQ_REGISTER_NODE,
                    cls = NodeRegistrationHandler,
                    current_requests = self.current_register_requests,
                    db = self.db,
                    devices = self.devices,
                    **self.kwargs)

    def show(self, requester, show_all):
        return show(self.db, requester, show_all)

    def get_node_info(self, requester, node_name):
        device_info = self.devices.get_device_info(requester, node_name)
        if device_info == None:
            return None # error already reported
        device_type = device_info.type
        if not is_a_node_type_name(device_type):
            requester.stderr.write('%s is not a node, it is a %s.\n' % \
                                    (node_name, device_type))
            return None
        node_info = self.db.select_unique("nodes", mac=device_info.mac)
        if node_info == None:
            requester.stderr.write(MSG_NODE_NEVER_REGISTERED % node_name)
            return None
        return merge_named_tuples(device_info, node_info)

    def get_reachable_node_info(self, requester, node_name, after_rescan = False):
        node_info = self.get_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        if node_info.reachable == 0:
            if after_rescan:
                requester.stderr.write(
                        'Connot reach %s. The node seems dead or disconnected.\n' % \
                                    node_name)
                return None
            else:
                # rescan, just in case, and retry
                self.devices.topology.rescan()   # just in case
                return self.get_reachable_node_info(
                        requester, node_name, after_rescan = True)
        return node_info

    def get_node_ip(self, requester, node_name):
        node_info = self.get_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        if node_info.ip == None:
            self.notify_unknown_ip(requester, node_name)
        return node_info.ip

    def get_reachable_nodes_ip(self, requester, node_set):
        nodes = self.parse_node_set(requester, node_set)
        result = []
        for node in nodes:
            node_info = self.get_reachable_node_info(requester, node.name)
            if node_info == None:
                return () # error already reported
            else:
                result.append(node_info.ip)
        return tuple(result)

    def filter_on_connectivity(self, requester, nodes, warn):
        nodes_ok = []
        nodes_ko = []
        for node in nodes:
            sw_ip, sw_port = self.devices.topology.get_connectivity_info( \
                                    node.mac)
            if sw_ip:
                nodes_ok.append(node)
            else:
                nodes_ko.append(node)
        if len(nodes_ko) > 0 and warn:
            requester.stderr.write(format_sentence_about_nodes(
                MSG_CONNECTIVITY_UNKNOWN, [n.name for n in nodes_ko]))
        return nodes_ok, nodes_ko

    def setpower(self, requester, node_set, poweron, warn_unknown_topology):
        nodes = self.parse_node_set(requester, node_set)
        if nodes == None:
            return None # error already reported
        # verify connectivity of all designated nodes
        for pass_count in range(2):
            nodes_ok, nodes_ko = self.filter_on_connectivity( \
                                requester, nodes,
                                pass_count > 0 and warn_unknown_topology)
            if len(nodes_ko) == 0:
                break
            elif pass_count == 0:
                # rescan and retry
                self.devices.topology.rescan()
        if len(nodes_ok) == 0:
            return False
        # otherwise, at least one node can be reached, so do it.
        for node in nodes_ok:
            self.devices.topology.setpower(node.mac, poweron)
        s_poweron = {True:'on',False:'off'}[poweron]
        requester.stdout.write(format_sentence_about_nodes(
            '%s was(were) powered ' + s_poweron + '.' ,
            [n.name for n in nodes_ok]) + '\n')
        return True

    def parse_node_set(self, requester, node_set):
        sql = None
        if node_set == None or node_set == 'my-nodes':
            sql = NODE_SET_QUERIES['my-nodes'] % requester.username
        elif node_set == 'all-nodes':
            sql = NODE_SET_QUERIES['all-nodes']
        if sql:
            nodes = [record[0] for record in self.db.execute(sql) ]
        else:
            # otherwise the list is explicit
            nodes = node_set.split(',')
        nodes = [self.get_node_info(requester, n) for n in nodes]
        if None in nodes:
            return None
        if len(nodes) == 0:
            requester.stderr.write('No matching nodes found! (tip: walt --help-about node-terminology)\n')
            return None
        return sorted(nodes)

    def wait(self, requester, q, node_set):
        nodes = self.parse_node_set(requester, node_set)
        self.wait_info.wait(requester, q, nodes)

    def node_bootup_event(self, node_name):
        node_info = self.get_node_info(None, node_name)
        self.wait_info.node_bootup_event(node_info)

    def includes_nodes_not_owned(self, requester, node_set, warn):
        nodes = self.parse_node_set(requester, node_set)
        if nodes == None:
            return None
        not_owned = [ n for n in nodes \
                if not (n.image.startswith(requester.username + '/') or
                        n.image.startswith('waltplatform/')) ]
        if len(not_owned) == 0:
            return False
        else:
            if warn:
                requester.stderr.write(format_sentence_about_nodes(
                    'Warning: %s seems(seem) to be used by another(other) user(users).',
                    [n.name for n in not_owned]) + '\n')
            return True

    def validate_cp(self, requester, src, dst):
        return validate_cp("node", self, requester, src, dst)

    def validate_cp_entity(self, requester, node_name):
        return self.get_reachable_node_ip(requester, node_name) != None

    def get_cp_entity_filesystem(self, requester, node_name):
        node_ip = self.get_node_ip(requester, node_name)
        return Filesystem(FS_CMD_PATTERN % dict(node_ip = node_ip))

    def get_cp_entity_attrs(self, requester, node_name):
        return dict(node_ip = self.get_node_ip(requester, node_name))



#!/usr/bin/env python
# coding=utf-8
import time, snimpy, sys, os
from walt.server import snmp, status, network
from walt.server.network.tools import lldp_update
from waltvlanconf import configure_main_switch_if_needed
from walt.common.tools import do
from walt.server.mydocker import DockerClient
from walt.server.const import WALT_INTF, SETUP_INTF, EXTERN_INTF

PLATFORM_READY_FILE = '/var/lib/walt/platform_ready'
MSG_WAITING_COMPANION_SERVICES = 'Waiting for companion services to be ready...'
MSG_DETECTING_MAIN_SWITCH = 'Detecting the main switch...'
MSG_AUTO_CONFIGURING_MAIN_SWITCH = 'Auto-configuring the main switch...'
MSG_ERROR_RESET_THE_SWITCH_REBOOT = '''\
Could not communicate with the main switch. Try to reset it to factory settings.
The server will reboot and retry this bootup procedure in a few seconds.
'''
MSG_DHCP_REQUEST = "Waiting for internet connection..."
MSG_CHECK_DOCKER = "Checking internet connection to the Docker Hub..."
MSG_WAITING_USER = "Waiting for user action..."

EXPLAIN_PLUG_SERVER = u'''\
We will first connect the WalT server and the main switch:

┌────────────────┐
│main switch     │
│1 2 [3-and-more]│
└┼─┼─┼─┼─┼─┼─┼─┼─┘
   │
   │
   └ walt server
'''

EXPLAIN_PLUG_ELEMENTS = u'''\
We will now connect the other elements of the WalT platform: cascaded switches and nodes.

┌────────────────┐      ┌─────────────┐
│main switch     │      │switch2      │
│1 2 [3-and-more]│      │[any port...]│
└┼─┼─┼─┼─┼─┼─┼─┼─┘      └┼─┼─┼─┼─┼─┼─┼┘
   │       │ └─────────────┘   │ └ node_Y
   │       └ node_X            └ node_Z
   └ walt server

You can use ports 3 and more of the main switch, and any port of the other switches
to connect these elements together.
Once a given switch is connected, power if off and on in order to ensure that its internal
network configuration is reset.
'''

EXPLAIN_PLUG_INTERNET_CABLE = u'''\
The cable to the external VLAN should now be plugged on port 1 of the main switch:

┌────────────────┐      ┌─────────────┐
│main switch     │      │switch2      │
│1 2 [3-and-more]│      │[any port...]│
└┼─┼─┼─┼─┼─┼─┼─┼─┘      └┼─┼─┼─┼─┼─┼─┼┘
 │ │       │ └─────────────┘   │ └ node_Y
 │ │       └ node_X            └ node_Z
 │ └ walt server
 └ to internet

This external VLAN is expected to provide a DHCP service and internet access (at least to the docker hub).
'''

ACTION_PLUG_SERVER = u'''
Please plug the WalT server to port number 2 of one of the switches (and power this switch on).
Do not plug anything more for now.
The procedure will continue when this switch is detected (this may take a few minutes even if it is already connected).
'''

ACTION_PLUG_ELEMENTS = u'''
Please plug the other switches and nodes as described above.
Press <Enter> when ready.
'''

ACTION_PLUG_INTERNET_CABLE = u'''
Please plug the cable of the external VLAN to port 1 of the main switch.
The server (mac address %(mac)s) will now be sending DHCP requests on this externally managed VLAN.
The setup process will continue when a DHCP offer is obtained.
'''

ACTION_CHECK_DOCKER = u'''
The server failed to access hub.docker.com. Please verify that the server is allowed to reach internet.
The setup process will retry this periodically until it succeeds.
'''

def main_switch_conf_callback(local_ip, switch_ip, intf, ui):
    ui.task_running()
    configured = configure_main_switch_if_needed(
                        snmp.Proxy(str(switch_ip), vlan=True))
    return configured

def wait_companion_services(ui):
    ui.task_start(MSG_WAITING_COMPANION_SERVICES)
    server_ip = str(network.tools.get_server_ip())
    while True:
        try:
            snmp_local = snmp.Proxy('localhost', lldp=True)
            print snmp_local.lldp.get_local_ips()
        except snimpy.snmp.SNMPNoSuchObject:
            ui.task_running()
            time.sleep(1)
        else:
            break
    ui.task_done()

def check_docker(ui):
    ui.task_start(  MSG_CHECK_DOCKER,
                    explain=EXPLAIN_PLUG_INTERNET_CABLE,
                    todo=ACTION_CHECK_DOCKER)
    docker = DockerClient()
    while not docker.self_test():
        ui.task_running()
        time.sleep(1)
    ui.task_done()

def setup(ui):
    wait_companion_services(ui)
    if not os.path.exists(PLATFORM_READY_FILE):
        setup_platform(ui)
        # create the file
        with open(PLATFORM_READY_FILE, 'a'):
            pass
    else:
        init_after_system_restart(ui)

def init_after_system_restart(ui):
    main_switch_info = wait_for_main_switch(ui, MSG_DETECTING_MAIN_SWITCH)
    network.tools.dhcp_wait_ip(EXTERN_INTF, ui, MSG_DHCP_REQUEST)

def wait_for_main_switch(ui, msg, explain = None, todo = None):
    ui.task_start(msg, explain=explain, todo=todo)
    # retrieve info about the main switch by
    # using the local lldp & snmp daemons
    snmp_local = snmp.Proxy('localhost', lldp=True)
    while len(snmp_local.lldp.get_neighbors()) == 0:
        ui.task_running()
        lldp_update()
        time.sleep(1)
    ui.task_done()
    main_switch_info = snmp_local.lldp.get_neighbors().values()[0]
    print 'main switch:', main_switch_info
    return main_switch_info

def auto_configure_main_switch(ui, switch_ip):
    ui.task_start(MSG_AUTO_CONFIGURING_MAIN_SWITCH)
    reached = False
    # assign a temporary ip in order to communicate
    # with the main switch, and set up its vlan configuration
    ui.task_running()
    reached, configured = network.tools.assign_temp_ip_to_reach_neighbor(
                                switch_ip,
                                main_switch_conf_callback,
                                SETUP_INTF,
                                ui)
    if not reached:
        ui.task_failed(MSG_ERROR_RESET_THE_SWITCH_REBOOT)
        time.sleep(10)
        do('reboot')
        time.sleep(10)
        sys.exit()
    ui.task_done()

def wait_other_elements_plugged(ui):
    ui.task_start(MSG_WAITING_USER, explain=EXPLAIN_PLUG_ELEMENTS,
                todo=ACTION_PLUG_ELEMENTS)
    ui.task_running(activity_sign = False)
    ui.wait_user_keypress()
    ui.task_done()

def setup_platform(ui):
    main_switch_info = wait_for_main_switch(
                    ui,
                    MSG_DETECTING_MAIN_SWITCH,
                    explain=EXPLAIN_PLUG_SERVER,
                    todo=ACTION_PLUG_SERVER)
    switch_ip = network.tools.ip(main_switch_info['ip'])
    auto_configure_main_switch(ui, switch_ip)
    wait_other_elements_plugged(ui)
    # get a server ip on eth0.169 (external VLAN) using DHCP
    mac_addr = network.tools.get_mac_address(EXTERN_INTF)
    network.tools.dhcp_wait_ip(
                EXTERN_INTF,
                ui,
                MSG_DHCP_REQUEST,
                EXPLAIN_PLUG_INTERNET_CABLE,
                ACTION_PLUG_INTERNET_CABLE % dict(mac = mac_addr))
    # check connection to the docker hub
    check_docker(ui)




from docker import Client
from walt.server.images.shell import ImageShellSessionStore
from walt.server.images.search import search
from walt.server.images.clone import clone
from walt.server.images.show import show
from walt.server.images.rename import rename
from walt.server.images.remove import remove
from walt.server.images.duplicate import duplicate
from walt.server.transfer import validate_cp
from walt.server.images.fixowner import fix_owner
from walt.server.images.store import NodeImageStore

# About terminology: See comment about it in image.py.

class NodeImageManager(object):
    def __init__(self, db, blocking_manager, dhcpd, docker):
        self.db = db
        self.blocking = blocking_manager
        self.dhcpd = dhcpd
        self.docker = docker
        self.store = NodeImageStore(self.docker, self.db)
        self.shells = ImageShellSessionStore(self.docker, self.store)
    def update(self):
        self.store.refresh()
        self.store.update_image_mounts()
    def search(self, requester, q, keyword):
        search(q, self.blocking, self.docker, requester, keyword)
    def clone(self, **kwargs):
        clone(blocking = self.blocking,
              docker = self.docker,
              image_store = self.store,
              **kwargs)
    def show(self, username):
        return show(self.store, username)
    def rename(self, requester, image_tag, new_tag):
        rename(self.store, self.docker, requester, image_tag, new_tag)
    def remove(self, requester, image_tag):
        remove(self.store, self.docker, requester, image_tag)
    def duplicate(self, requester, image_tag, new_tag):
        duplicate(self.store, self.docker, requester, image_tag, new_tag)
    def validate_cp(self, requester, src, dst):
        return validate_cp("image", self, requester, src, dst)
    def validate_cp_entity(self, requester, image_tag):
        return self.has_image(requester, image_tag)
    def get_cp_entity_filesystem(self, requester, image_tag):
        return self.store.get_user_image_from_tag(requester, image_tag).filesystem
    def get_cp_entity_attrs(self, requester, image_tag):
        return dict(image_tag=image_tag)
    def fix_owner(self, requester, other_user):
        fix_owner(self.store, self.docker, requester, other_user)
    def cleanup(self):
        # give up image shell sessions
        self.shells.cleanup()
        # un-mount images
        self.store.cleanup()
    def has_image(self, requester, image_tag):
        return self.store.get_user_image_from_tag(requester, image_tag) != None
    def set_image(self, requester, node_macs, image_tag):
        image = self.store.get_user_image_from_tag(requester, image_tag)
        if image:
            for node_mac in node_macs:
                self.db.update('nodes', 'mac',
                        mac=node_mac,
                        image=image.fullname)
            self.store.update_image_mounts()
            self.db.commit()
            self.dhcpd.update()
        return image != None
    def create_shell_session(self, requester, image_tag):
        return self.shells.create_session(requester, image_tag)




import requests
from walt.server.tools import columnate, \
                display_transient_label, hide_transient_label

# About terminology: See comment about it in image.py.

LOCATION_WALT_SERVER = 0
LOCATION_DOCKER_HUB = 1
LOCATION_LABEL = {
    LOCATION_WALT_SERVER: 'server',
    LOCATION_DOCKER_HUB: 'hub',
}
LOCATION_LONG_LABEL = {
    LOCATION_WALT_SERVER: 'WalT server',
    LOCATION_DOCKER_HUB: 'docker hub',
}

LOCATION_PER_LABEL = {v: k for k, v in LOCATION_LABEL.items()}

class Search(object):
    def __init__(self, docker, requester, transient_label='Searching...'):
        self.docker = docker
        self.requester = requester
        self.result = {}
        self.transient_label = transient_label
    # search returns a dictionary with the following format:
    # { <tag> -> { <user> -> <location> } }
    def search(self, validate = None):
        display_transient_label(self.requester.stdout, self.transient_label)
        candidates = []
        if not validate:
            def validate(user, tag, location):
                return True
        # look up for candidates on the docker hub
        for result in self.docker.search('walt-node'):
            if '/walt-node' in result['name']:
                for user, tag in self.docker.lookup_remote_tags(result['name']):
                    candidates.append((user, tag, LOCATION_DOCKER_HUB))
        # look up for candidates locally on the server
        for fullname in self.docker.get_local_images():
            if '/walt-node' in fullname:
                user, tag = fullname.split('/walt-node:')
                candidates.append((user, tag, LOCATION_WALT_SERVER))
        # validate candidates
        for user, tag, location in candidates:
            if validate(user, tag, location):
                self.insert_result(tag, user, location)
        hide_transient_label(self.requester.stdout, self.transient_label)
        return self.result
    def insert_result(self, tag, user, location):
        if tag not in self.result:
            self.result[tag] = {}
        if user not in self.result[tag]:
            self.result[tag][user] = set([])
        self.result[tag][user].add(location)

# this implements walt image search
def perform_search(docker, requester, keyword):
    # images owned by the requester and present locally on
    # the server are not considered "remote images".
    # (they belong to the working set of the user, instead.)
    def validate_not_in_ws(user, tag, location):
        return user != requester.username or \
                location == LOCATION_DOCKER_HUB
    if keyword:
        def validate(user, tag, location):
            if not validate_not_in_ws(user, tag, location):
                return False
            remote_name = "%s:%s/%s" % ( \
                    LOCATION_LABEL[location], user, tag)
            return keyword in remote_name
    else:
        validate = validate_not_in_ws
    # search
    result = Search(docker, requester).search(validate)
    # print
    records = []
    for tag in result:
        for user in result[tag]:
            for location in result[tag][user]:
                clonable_link = "%s:%s/%s" % (\
                    LOCATION_LABEL[location],
                    user,
                    tag
                )
                records.append([
                    user, tag, LOCATION_LONG_LABEL[location], clonable_link
                ])
    return columnate(sorted(records), \
               ['User', 'Image name', 'Location', 'Clonable link'])

class SearchTask(object):
    def __init__(self, q, docker, requester, keyword):
        self.response_q = q
        self.docker = docker
        self.requester = requester
        self.keyword = keyword
    def perform(self):
        return perform_search(self.docker, self.requester, self.keyword)
    def handle_result(self, res):
        if isinstance(res, requests.exceptions.RequestException):
            res = 'Network connection to docker hub failed.'
        elif isinstance(res, Exception):
            raise res   # unexpected
        self.response_q.put(res)

# this implements walt image search
def search(q, blocking_manager, docker, requester, keyword):
    blocking_manager.do(SearchTask(q, docker, requester, keyword))





#!/usr/bin/env python
from walt.server.postgres import PostgresDB
from time import time

EV_AUTO_COMMIT              = 0
EV_AUTO_COMMIT_PERIOD       = 2

class ServerDB(PostgresDB):

    def __init__(self):
        # parent constructor
        PostgresDB.__init__(self)
        # create the db schema
        self.execute("""CREATE TABLE IF NOT EXISTS devices (
                    mac TEXT PRIMARY KEY,
                    ip TEXT,
                    name TEXT,
                    reachable INTEGER,
                    type TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS topology (
                    mac TEXT REFERENCES devices(mac),
                    switch_mac TEXT REFERENCES devices(mac),
                    switch_port INTEGER);""")
        self.execute("""CREATE TABLE IF NOT EXISTS images (
                    fullname TEXT PRIMARY KEY,
                    ready BOOLEAN);""")
        self.execute("""CREATE TABLE IF NOT EXISTS nodes (
                    mac TEXT REFERENCES devices(mac),
                    image TEXT REFERENCES images(fullname));""")
        self.execute("""CREATE TABLE IF NOT EXISTS config (
                    item TEXT PRIMARY KEY,
                    value TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS logstreams (
                    id SERIAL PRIMARY KEY,
                    sender_mac TEXT REFERENCES devices(mac),
                    name TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS logs (
                    stream_id INTEGER REFERENCES logstreams(id),
                    timestamp TIMESTAMP,
                    line TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS checkpoints (
                    username TEXT,
                    timestamp TIMESTAMP,
                    name TEXT);""")

    # Some types of events are numerous and commiting the
    # database each time would be costly.
    # That's why we auto-commit every few seconds.
    def plan_auto_commit(self, ev_loop):
        ev_loop.plan_event(
            ts = time(),
            target = self,
            repeat_delay = EV_AUTO_COMMIT_PERIOD,
            ev_type = EV_AUTO_COMMIT
        )

    def handle_planned_event(self, ev_type):
        assert(ev_type == EV_AUTO_COMMIT)
        self.commit()

    def get_logs(self, c, **kwargs):
        sql, args = self.format_logs_query('l.*', ordering='l.timestamp', **kwargs)
        c.execute(sql, args)
        return c

    def count_logs(self, **kwargs):
        sql, args = self.format_logs_query('count(*)', **kwargs)
        return self.execute(sql, args).fetchall()[0][0]

    def format_logs_query(self, projections, ordering=None, \
                    senders=None, history=(None,None), **kwargs):
        args = []
        constraints = [ 's.sender_mac = d.mac', 'l.stream_id = s.id' ]
        if senders:
            constraints.append('''d.name IN ('%s') ''' % "','".join(senders))
        start, end = history
        if start:
            constraints.append('l.timestamp > %s')
            args.append(start)
        if end:
            constraints.append('l.timestamp < %s')
            args.append(end)
        where_clause = self.get_where_clause_from_constraints(constraints)
        if ordering:
            ordering = 'order by ' + ordering
        else:
            ordering = ''
        return ("SELECT %s FROM devices d, logstreams s, logs l %s %s;" % \
                                (projections, where_clause, ordering), args)

    def forget_device(self, dev_name):
        self.execute("""
            DELETE FROM logs l USING devices d, logstreams s
                WHERE d.name = %s AND s.sender_mac = d.mac AND l.stream_id = s.id;
            DELETE FROM logstreams s USING devices d WHERE d.name = %s AND s.sender_mac = d.mac;
            DELETE FROM nodes n USING devices d WHERE d.name = %s AND d.mac = n.mac;
            DELETE FROM topology t USING devices d WHERE d.name = %s AND d.mac = t.mac;
            DELETE FROM topology t USING devices d WHERE d.name = %s AND d.mac = t.switch_mac;
            DELETE FROM devices d WHERE d.name = %s;
        """,  (dev_name,)*6)
        self.commit()

    def get_config(self, item, default = None):
        res = self.select_unique("config", item=item)
        if res == None:
            if default == None:
                raise RuntimeError(\
                    "Failed get_config(): item not found and no default provided.")
            self.insert("config", item=item, value=default)
            self.commit()
            return default
        else:
            return res.value

    def set_config(self, item, value):
        res = self.select_unique("config", item=item)
        if res == None:
            self.insert("config", item=item, value=value)
        else:
            self.update("config", "item", item=item, value=value)



import os, re, time, shutil
from plumbum.cmd import mount, umount, chroot
from walt.server.network.tools import get_server_ip
from walt.server.filesystem import Filesystem
from walt.common.tools import \
        failsafe_makedirs, succeeds

# IMPORTANT TERMINOLOGY NOTES:
#
# In the client side source code
# ------------------------------
# we follow the *walt user point of view*:
# the 'name' of an image is actually the docker <tag> only.
#
# In the server side source code
# ------------------------------
# we follow the *docker terminology*
#
# <user>/<repository>:<tag>
# <-- name --------->
# <-- fullname ----------->
#
# note: all walt images have <repository>='walt-node',
# which allows to find them. (1)
# As a result of (1), all docker images that walt manages
# have a fullname with the following format:
# <username>/walt-node:<tag>
#
# About clonable images
# ---------------------
# For a given user <username>=<u>,
# the usage of an image is describe by the table below:
#
# In the docker hub  | yes | yes | yes | yes | no  | no  |
# On the walt server | yes | yes | no  | no  | yes | yes |
# User is <u>        | yes | no  | yes | no  | yes | no  |
# --------------------------------------------------------
# displayed by show  | yes | no  | no  | no  | yes | no  |
# returned by search | no  | yes | yes | yes | no  | yes |
# <u> may clone it   | no  | yes | yes | yes | no  | yes |
#
# In order to clone an image, the user must specify
# a 'clonable image link' as an argument to 'walt image clone'.
# Such a clonable link is formatted as follows:
# [server|hub]:<user>/<image_tag>

IMAGE_MOUNT_PATH='/var/lib/walt/images/%s/fs'
IMAGE_DIFF_PATH='/var/lib/walt/images/%s/diff'
SERVER_PUBKEY_PATH = '/root/.ssh/id_dsa.pub'
HOSTS_FILE_CONTENT="""\
127.0.0.1   localhost
::1     localhost ip6-localhost ip6-loopback
ff02::1     ip6-allnodes
ff02::2     ip6-allrouters
"""

def get_mount_path(image_fullname):
    mount_path, dummy = get_mount_info(image_fullname)
    return mount_path

def get_mount_info(image_fullname):
    sub_path = re.sub('[^a-zA-Z0-9/-]+', '_', re.sub(':', '/', image_fullname))
    return IMAGE_MOUNT_PATH % sub_path, IMAGE_DIFF_PATH % sub_path

def get_server_pubkey():
    with open(SERVER_PUBKEY_PATH, 'r') as f:
        return f.read()

def parse_image_fullname(image_fullname):
    image_name, image_tag = image_fullname.split(':')
    image_user, image_repo = image_name.split('/')
    return image_fullname, image_name, image_repo, image_user, image_tag

def validate_image_tag(requester, image_tag):
    is_ok = re.match('^[a-zA-Z0-9\-]+$', image_tag)
    if not is_ok:
        requester.stderr.write(\
                'Bad name: Only alnum and dash(-) characters are allowed.\n')
    return is_ok

FS_CMD_PATTERN = 'docker run --rm --entrypoint %%(prog)s %(image)s %%(prog_args)s'

class NodeImage(object):
    server_pubkey = get_server_pubkey()
    def __init__(self, docker, fullname):
        self.docker = docker
        self.rename(fullname)
        self.created_at = None
        self.ready = False
        self.mount_path = None
        self.mounted = False
        self.top_layer_id = None
        self.server_ip = get_server_ip()
        self.filesystem = Filesystem(FS_CMD_PATTERN % dict(image = self.fullname))
    def rename(self, fullname):
        self.fullname, self.name, dummy, self.user, self.tag = \
            parse_image_fullname(fullname)
    def set_ready(self, is_ready):
        if is_ready and not self.ready:
            # image just became ready, get the creation time and image id from docker
            self.created_at = self.docker.get_creation_time(self.fullname)
            self.update_top_layer_id()
        self.ready = is_ready
    def update_top_layer_id(self):
        self.top_layer_id = self.docker.get_top_layer_id(self.fullname)
    def __del__(self):
        if self.mounted:
            self.unmount()
    def get_mount_info(self):
        return get_mount_info(self.fullname)
    def mount(self):
        print 'Mounting %s...' % self.fullname,
        self.mount_path, self.diff_path = self.get_mount_info()
        failsafe_makedirs(self.mount_path)
        failsafe_makedirs(self.diff_path)
        layers = self.docker.get_image_layers(self.top_layer_id)
        branches = [ layer + '=ro' for layer in layers ]
        branches.insert(0, self.diff_path + '=rw')
        branches_opt = 'br=' + ':'.join(branches)
        mount('-t', 'aufs', '-o', branches_opt, 'none', self.mount_path)
        chroot(self.mount_path, 'walt-node-install', self.server_ip, NodeImage.server_pubkey)
        self.create_hosts_file()
        self.mounted = True
        print 'done'
    def unmount(self):
        print 'Un-mounting %s...' % self.fullname,
        while not succeeds('umount %s 2>/dev/null' % self.mount_path):
            time.sleep(0.1)
        shutil.rmtree(self.diff_path)
        os.rmdir(self.mount_path)
        self.mounted = False
        print 'done'
    def create_hosts_file(self):
        # since file /etc/hosts is managed by docker,
        # it appears empty on the bind mount.
        # let's create it appropriately.
        with open(self.mount_path + '/etc/hosts', 'w') as f:
            f.write(HOSTS_FILE_CONTENT)




from walt.server.tools import columnate

MSG_WS_IS_EMPTY="""\
Your working set is empty.
Use 'walt image search [<keyword>]' to search for images
you could build upon.
Then use 'walt image clone <clonable_link>' to clone them
into your working set.
"""

def show(images, username):
    tabular_data = []
    for fullname, image in images.iteritems():
        if image.user != username:
            continue
        tabular_data.append([
                    image.tag,
                    str(image.mounted),
                    image.created_at if image.created_at else 'N/A',
                    str(image.ready) ])
    if len(tabular_data) > 0:
        header = [ 'Name', 'Mounted', 'Created', 'Ready' ]
        return columnate(tabular_data, header)
    else:
        return MSG_WS_IS_EMPTY
